// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef DUNE_CURVED_SURFACE_GRID_GRIDFAMILY_HH
#define DUNE_CURVED_SURFACE_GRID_GRIDFAMILY_HH

#include <type_traits>

#include <dune/grid/common/grid.hh>
#include <dune/curvedsurfacegrid/capabilities.hh>
#include <dune/curvedsurfacegrid/entity.hh>
#include <dune/curvedsurfacegrid/entityseed.hh>
#include <dune/curvedsurfacegrid/geometry.hh>
#include <dune/curvedsurfacegrid/gridview.hh>
#include <dune/curvedsurfacegrid/intersection.hh>
#include <dune/curvedsurfacegrid/intersectioniterator.hh>
#include <dune/curvedsurfacegrid/iterator.hh>
#include <dune/curvedsurfacegrid/idset.hh>
#include <dune/curvedsurfacegrid/indexsets.hh>
#include <dune/curvedsurfacegrid/gridfunctions/gridfunction.hh>

#include <dune/functions/common/functionconcepts.hh>
#include <dune/functions/common/signature.hh>
#include <dune/functions/gridfunctions/localderivativetraits.hh>

namespace Dune
{

  /** \brief namespace containing the implementations of CurvedSurfaceGrid
   *  \ingroup CurvedGeo
   */
  namespace CGeo
  {
    template< class GF >
    struct DimRange
    {
      using SigTraits = Functions::SignatureTraits<GF>;
      using Range = typename SigTraits::RawRange;
      static const int value = Range::size();
    };

    template< class GF, class LF >
    struct DifferentiableLocalFunction
    {
      using EntitySet = typename GF::EntitySet;
      using LocalContext = typename EntitySet::Element;

      using Range = typename Functions::SignatureTraits<GF>::Range;
      using LocalSignature = Range(typename EntitySet::LocalCoordinate);

      static const bool value = Functions::Concept::isDifferentiableLocalFunction<LF,LocalSignature,LocalContext,Functions::LocalDerivativeTraits<EntitySet>::template Traits>();
    };

    // GridFamily
    // ----------

    template< class GF, int order >
    struct GridFamily
    {
      struct Traits
      {
        using Grid = CurvedSurfaceGrid<GF,order>;
        using HostGrid = GridOf_t<GF>;

        using GridFunction = GF;
        using LocalFunction = std::decay_t<decltype(localFunction(std::declval<GF const&>()))>;

        static const bool differentiableLocalFunction = DifferentiableLocalFunction<GF, LocalFunction>::value;
        static_assert((differentiableLocalFunction || order>0), "Either provide a differentiable GridFunction or set ORDER > 0");

        using ctype = typename HostGrid::ctype;

        static const int dimension = HostGrid::dimension;
        static const int dimensionworld = DimRange<GridFunction>::value;

        using LeafIntersection
          = Dune::Intersection<const Grid, CGeo::Intersection<const Grid, typename HostGrid::LeafIntersection>>;
        using LevelIntersection
          = Dune::Intersection<const Grid, CGeo::Intersection<const Grid, typename HostGrid::LevelIntersection>>;

        using LeafIntersectionIterator
          = Dune::IntersectionIterator<const Grid, CGeo::IntersectionIterator<const Grid, typename HostGrid::LeafIntersectionIterator>, CGeo::Intersection<const Grid, typename HostGrid::LeafIntersection> >;
        using LevelIntersectionIterator
          = Dune::IntersectionIterator<const Grid, CGeo::IntersectionIterator<const Grid, typename HostGrid::LevelIntersectionIterator >, CGeo::Intersection<const Grid, typename HostGrid::LevelIntersection> >;

        using HierarchicIterator
          = Dune::EntityIterator<0, const Grid, CGeo::HierarchicIterator<const Grid>>;

        template< int codim >
        struct Codim
        {
          using LocalGeometry = typename HostGrid::template Codim<codim>::LocalGeometry;

          using LocalTransformation = std::conditional_t<(codim == 0),
            CGeo::Impl::DefaultLocalGeometry,
            CGeo::Impl::LocalGeometryInterface<LocalGeometry>>;

          template < int mydim, int cdim, class GridImpl >
          using GeometryImplTemplate
            = CGeo::Geometry<ctype, mydim, cdim, LocalFunction, LocalTransformation, order>;

          // geometry types
          using GeometryImpl = GeometryImplTemplate<dimension-codim, dimensionworld, const Grid>;
          using Geometry = Dune::Geometry<dimension-codim, dimensionworld, const Grid, GeometryImplTemplate>;

          // entity types
          using EntityImpl = CGeo::Entity<codim, dimension, const Grid>;
          using Entity = Dune::Entity<codim, dimension, const Grid, CGeo::Entity>;

          using EntitySeed = Dune::EntitySeed<const Grid, CGeo::EntitySeed<codim, const Grid> >;

          template< PartitionIteratorType pitype >
          struct Partition
          {
            using LeafIteratorImp
              = CGeo::Iterator<typename HostGrid::LeafGridView, codim, pitype, const Grid>;
            using LeafIterator = Dune::EntityIterator<codim, const Grid, LeafIteratorImp>;

            using LevelIteratorImp
              = CGeo::Iterator<typename HostGrid::LevelGridView, codim, pitype, const Grid>;
            using LevelIterator = Dune::EntityIterator<codim, const Grid, LevelIteratorImp>;
          };

          using LeafIterator = typename Partition< All_Partition >::LeafIterator;
          using LevelIterator = typename Partition< All_Partition >::LevelIterator;
        };

        // index-sets
        using LeafIndexSet = CGeo::IndexSet<const Grid, typename HostGrid::Traits::LeafIndexSet>;
        using LevelIndexSet = CGeo::IndexSet<const Grid, typename HostGrid::Traits::LevelIndexSet>;

        // id-sets
        using GlobalIdSet = CGeo::IdSet<const Grid, typename HostGrid::Traits::GlobalIdSet>;
        using LocalIdSet = CGeo::IdSet<const Grid, typename HostGrid::Traits::LocalIdSet>;

        using CollectiveCommunication = typename HostGrid::Traits::CollectiveCommunication;

        // grid views
        using LeafGridView
          = Dune::GridView<CGeo::GridViewTraits<typename HostGrid::LeafGridView, GF, order>>;
        using LevelGridView
          = Dune::GridView<CGeo::GridViewTraits<typename HostGrid::LevelGridView, GF, order>>;
      };
    };

  } // namespace CGeo
} // namespace Dune

#endif // DUNE_CURVED_SURFACE_GRID_GRIDFAMILY_HH
