// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef DUNE_CURVED_SURFACE_GRID_SPHERE_GRIDFUNCTION_HH
#define DUNE_CURVED_SURFACE_GRID_SPHERE_GRIDFUNCTION_HH

#include <type_traits>

#include <dune/curvedsurfacegrid/gridfunctions/analyticgridfunction.hh>
#include <dune/functions/common/defaultderivativetraits.hh>

namespace Dune
{
  // Sphere functor
  template< class T >
  struct SphereProjection
  {
    T radius_;

    SphereProjection (T radius)
      : radius_(radius)
    {}

    //! project the coordinate to the sphere at origin with `radius`
    template< class Domain >
    Domain operator() (const Domain& x) const
    {
      return x * (radius_ / x.two_norm());
    }

    //! derivative of the projection
    friend auto derivative (const SphereProjection& sphere)
    {
      return [r=sphere.radius_](auto const& x)
      {
        using Domain = std::decay_t<decltype(x)>;
        using DerivativeTraits = Functions::DefaultDerivativeTraits<Domain(Domain)>;
        typename DerivativeTraits::Range out;

        auto nrm = x.two_norm();
        for (int i = 0; i < out.N(); ++i)
          for (int j = 0; j < out.M(); ++j)
            out[i][j] = r * ((i == j ? 1 : 0) - (x[i]/nrm) * (x[j]/nrm)) / nrm;
        return out;
      };
    }
  };

  //! construct a grid function representing a sphere parametrization
  template< class Grid, class T >
  auto sphereGridFunction (T radius)
  {
    return analyticGridFunction<Grid>(SphereProjection<T>{radius});
  }

} // end namespace Dune

#endif // DUNE_CURVED_SURFACE_GRID_SPHERE_GRIDFUNCTION_HH
