// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef DUNE_CURVED_SURFACE_GRID_ITERATOR_HH
#define DUNE_CURVED_SURFACE_GRID_ITERATOR_HH

#include <cassert>
#include <type_traits>
#include <utility>

#include <dune/geometry/referenceelements.hh>
#include <dune/curvedsurfacegrid/entity.hh>

namespace Dune
{
  namespace CGeo
  {
    // Internal Forward Declarations
    // -----------------------------

    template< class HostGridView, int codim, PartitionIteratorType pitype, class Grid >
    class Iterator;

    template< class Grid >
    class HierarchicIterator;


    // Iterator
    // --------

    template< class HostGridView, int codim, PartitionIteratorType pitype, class G >
    class Iterator
    {
      using Traits = typename std::remove_const_t<G>::Traits;

    public:
      static const int codimension = codim;

      using Grid = typename Traits::Grid;
      using GridFunction = typename Traits::GridFunction;
      using Entity = typename Traits::template Codim<codimension>::Entity;

    private:
      using EntityImpl = CGeo::Entity<codimension, Traits::dimension, G>;
      using HostEntityIterator = typename HostGridView::template Codim<codim>::template Partition<pitype>::Iterator;

    public:
      Iterator () = default;

      Iterator (const GridFunction& gridFunction, HostEntityIterator hostEntityIterator)
        : gridFunction_(&gridFunction)
        , hostEntityIterator_(std::move(hostEntityIterator))
      {}

      void increment ()
      {
        ++hostEntityIterator_;
      }

      bool equals (const Iterator& rhs) const
      {
        return hostEntityIterator_ == rhs.hostEntityIterator_;
      }

      Entity dereference () const
      {
        return EntityImpl(gridFunction(), *hostEntityIterator_);
      }

      int level () const
      {
        return hostEntityIterator_.level();
      }

      const GridFunction& gridFunction () const
      {
        return *gridFunction_;
      }

      static Iterator begin (const GridFunction& gf, const HostGridView& hostGridView)
      {
        HostEntityIterator hostEntityIterator = hostGridView.template begin<codimension, pitype>();
        return Iterator(gf, std::move(hostEntityIterator));
      }

      static Iterator end (const GridFunction& gf, const HostGridView& hostGridView)
      {
        HostEntityIterator hostEntityIterator = hostGridView.template end<codimension, pitype>();
        return Iterator(gf, std::move(hostEntityIterator));
      }

    private:
      const GridFunction* gridFunction_;
      HostEntityIterator hostEntityIterator_;
    };


    // HierarchicIterator
    // ------------------

    template< class G >
    class HierarchicIterator
    {
      using Traits = typename std::remove_const_t<G>::Traits;

    public:
      static const int codimension = 0;

      using Grid = typename Traits::Grid;
      using GridFunction = typename Traits::GridFunction;
      using Entity = typename Traits::template Codim<codimension>::Entity;

    private:
      using EntityImpl = CGeo::Entity<codimension, Traits::dimension, G>;
      using HostEntityIterator = typename Grid::Traits::HostGrid::HierarchicIterator;

    public:
      HierarchicIterator () = default;

      //! Constructor. Stores a pointer to the grid and the iterator by value.
      HierarchicIterator (const GridFunction& gridFunction, HostEntityIterator hostEntityIterator)
        : gridFunction_(&gridFunction)
        , hostEntityIterator_(std::move(hostEntityIterator))
      {}

      void increment ()
      {
        ++hostEntityIterator_;
      }

      bool equals (const HierarchicIterator& rhs) const
      {
        return hostEntityIterator_ == rhs.hostEntityIterator_;
      }

      Entity dereference () const
      {
        return EntityImpl(gridFunction(), *hostEntityIterator_);
      }

      int level () const
      {
        return hostEntityIterator_.level();
      }

      const GridFunction& gridFunction () const
      {
        return *gridFunction_;
      }

    private:
      const GridFunction* gridFunction_ = nullptr;
      HostEntityIterator hostEntityIterator_;
    };

  } // namespace CGeo
} // namespace Dune

#endif // DUNE_CURVED_SURFACE_GRID_ITERATOR_HH
