#ifndef DUNE_CURVED_SURFACE_GRID_DATASTRUCTURES_HH
#define DUNE_CURVED_SURFACE_GRID_DATASTRUCTURES_HH

#include <fstream>
#include <sstream>
#include <vector>
#include <list>
#include <stdint.h>

namespace Dune{ namespace CGeo{

const int output_precision = 10;
const size_t npos = std::string::npos;
const double epsilon = 1e-9; //precision of floating point operations
const double coarse_epsilon = 1e-4;

typedef unsigned char uchar;

//--------------------< CurvedSurfaceGridException >
struct CurvedSurfaceGridException: public std::exception{
  std::string msg;
  CurvedSurfaceGridException(const std::string &msg):msg(msg){/*$ACL_exclude*/}
  virtual ~CurvedSurfaceGridException() throw(){/*$ACL_exclude*/}
  virtual const char* what() const throw(){/*$ACL_exclude*/return msg.c_str();}
};

//--------------------< error >
inline void error(std::string message){ //$ACL_exclude
  throw CurvedSurfaceGridException("ERROR! " + message);
}

//--------------------< ft (int32_t) >
inline std::string ft(int32_t value, unsigned char minlen=0, char fill_char='0'){ //$ACL_exclude
  std::ostringstream convert;
  convert << value;
  std::string tmp = convert.str();
  int required_zeros = minlen - tmp.size();
  if(required_zeros > 0){
    if(value >= 0) tmp = std::string(required_zeros, fill_char) + tmp;
    else if(fill_char == '0') tmp = "-" + std::string(required_zeros, fill_char) + tmp.substr(1);
    else tmp = std::string(required_zeros, fill_char) + "-" + tmp.substr(1);
  }
  return tmp;
}

//--------------------< ft (uint32_t) >
inline std::string ft(uint32_t value, unsigned char minlen=0, char fill_char='0'){ //$ACL_exclude
  std::ostringstream convert;
  convert << value;
  std::string tmp = convert.str();
  int required_zeros = minlen - tmp.size();
  if(required_zeros > 0) tmp = std::string(required_zeros, fill_char) + tmp;
  return tmp;
}

//--------------------< ft (double) >
inline std::string ft(double value, unsigned char minlen=0, char tail=-1, char fill_char='0'){
  //$ACL_exclude
  std::ostringstream convert;
  convert.precision(output_precision);
  //convert.setf(ios_base::fixed);
  convert << value;
  std::string tmp = convert.str();
  size_t tail_size, pos = tmp.rfind(".");
  if(pos == npos){
    if(tail > 0) tmp += ".";
    tail_size = 0;
  }else{
    tail_size = tmp.size() - pos - 1;
  }
  int required_zeros;
  if(tail >= 0){
    required_zeros = tail - tail_size;
    if(required_zeros > 0) tmp += std::string(required_zeros, '0');
    else if(required_zeros < 0) tmp = tmp.substr(0, tmp.size() + required_zeros);
  }
  if(tmp[tmp.size() - 1] == '.') tmp = tmp.substr(0, tmp.size() - 1);
  required_zeros = minlen - tmp.size();
  if(required_zeros > 0){
    if(value >= 0) tmp = std::string(required_zeros, fill_char) + tmp;
    else tmp = "-" + std::string(required_zeros, fill_char) + tmp.substr(1);
  }
  return tmp;
}

//--------------------< ival >
inline int ival(std::string str){ //$ACL_exclude
  int result;
  std::istringstream convert(str);
  if(!(convert >> result)) result = 0;
  return result;
}

//--------------------< ival64 >
inline int64_t ival64(std::string str){ //$ACL_exclude
  int64_t result;
  std::istringstream convert(str);
  if(!(convert >> result)) result = 0;
  return result;
}

//--------------------< dval >
inline double dval(std::string str){ //$ACL_exclude
  double result;
  std::istringstream convert(str);
  if(!(convert >> result)) result = 0.0;
  return result;
}

//--------------------< ltrim >
inline std::string ltrim(std::string str){ //$ACL_exclude
  if(str.size() == 0) return "";
  unsigned i = 0;
  while(isspace(str[i]) && i < str.size() - 1) ++i;
  return str.substr(i);
}

//--------------------< DataVector >
typedef std::vector<double> DataVector;

//--------------------< Vertex >
struct Vertex{
  double c[3];

  Vertex(){/*$ACL_exclude*/c[0]=0.0; c[1]=0.0; c[2]=0.0;}
  explicit Vertex(double x, double y=0.0, double z=0.0){/*$ACL_exclude*/c[0]=x; c[1]=y; c[2]=z;}
  Vertex(const Vertex &v){/*$ACL_exclude*/c[0]=v[0]; c[1]=v[1]; c[2]=v[2];}
  const double& operator[](const int index) const {/*$ACL_exclude*/return c[index];}
  double& operator[](const int index){/*$ACL_exclude*/return c[index];}
  Vertex& operator+=(const Vertex &rhs);
  Vertex& operator+=(const double f);
  Vertex& operator-=(const Vertex &rhs);
  Vertex& operator-=(const double f);
  Vertex& operator*=(const double f);
  Vertex& operator/=(const double f);
  friend std::ostream& operator<<(std::ostream &os, const Vertex &v){ //$ACL_exclude
                        os << "(" << v.c[0] << ", " << v.c[1] << ", " << v.c[2] << ")"; return os;}
  std::string print(){ //$ACL_exclude
                      return "(" + ft(c[0]) + ", " + ft(c[1]) + ", " + ft(c[2]) + ")";}
  bool is_origin(){/*$ACL_exclude*/return (c[0] == 0.0 && c[1] == 0.0 && c[2] == 0.0);}
};

inline Vertex operator+(const Vertex &v1, const Vertex &v2){ //$ACL_exclude
  return Vertex(v1.c[0] + v2.c[0], v1.c[1] + v2.c[1], v1.c[2] + v2.c[2]);
}

inline Vertex operator+(const Vertex &v1, const double f){ //$ACL_exclude
  return Vertex(v1.c[0] + f, v1.c[1] + f, v1.c[2] + f);
}

inline Vertex operator-(const Vertex &v1, const Vertex &v2){ //$ACL_exclude
  return Vertex(v1.c[0] - v2.c[0], v1.c[1] - v2.c[1], v1.c[2] - v2.c[2]);
}

inline Vertex operator-(const Vertex &v1, const double f){ //$ACL_exclude
  return Vertex(v1.c[0] - f, v1.c[1] - f, v1.c[2] - f);
}

inline Vertex operator*(const Vertex &v, const double f){ //$ACL_exclude
  return Vertex(v.c[0] * f, v.c[1] * f, v.c[2] * f);
}

inline Vertex operator/(const Vertex &v, const double f){ //$ACL_exclude
  return Vertex(v.c[0] / f, v.c[1] / f, v.c[2] / f);
}

inline Vertex& Vertex::operator+=(const Vertex &rhs){ //$ACL_exclude
  *this = *this + rhs;
  return *this;
}

inline Vertex& Vertex::operator+=(const double f){ //$ACL_exclude
  *this = *this + f;
  return *this;
}

inline Vertex& Vertex::operator-=(const Vertex &rhs){ //$ACL_exclude
  *this = *this - rhs;
  return *this;
}

inline Vertex& Vertex::operator-=(const double f){ //$ACL_exclude
  *this = *this - f;
  return *this;
}

inline Vertex& Vertex::operator*=(const double f){ //$ACL_exclude
  *this = *this * f;
  return *this;
}

inline Vertex& Vertex::operator/=(const double f){ //$ACL_exclude
  *this = *this / f;
  return *this;
}

inline std::string ft(Vertex value, unsigned char minlen, char tail, char fill_char){ //$ACL_exclude
  return "(" + ft(value[0], minlen, tail, fill_char) + ", " + ft(value[1], minlen, tail, fill_char)
                                               + ", " + ft(value[2], minlen, tail, fill_char) + ")";
}

//--------------------< Element >
struct Element{
  int v[3]; //vertex-indices
  int operation_index;

  friend std::ostream& operator<<(std::ostream &os, const Element &e){/*$ACL_exclude*/
        os << "(" << e.v[0] << ", " << e.v[1] << ", " << e.v[2] << ", " << e.v[3] <<")"; return os;}
  std::string print(){/*$ACL_exclude*/ return "(" + ft(v[0]) + ", " + ft(v[1]) + ", "
                                                                + ft(v[2]) + ", " + ft(v[3]) + ")";}
};

//--------------------< Mesh >
struct Mesh{
  std::vector<Vertex> vertices;
  std::list<Element> elements;
  short dim_of_elements;
  short dim_of_world;

  //constructors
  Mesh(int dim_of_elements=2, int dim_of_world=3):dim_of_elements(dim_of_elements),
                                        dim_of_world(dim_of_world){/*$ACL_exclude*/}

  //add_vertex
  void add_vertex(const Vertex &v){ //$ACL_exclude
    vertices.push_back(v);
    if(dim_of_world == 2) vertices.back()[2] = 0.0;
  }

  //add_element
  void add_element(const int32_t *v){ //$ACL_exclude
    elements.push_back(Element());
    Element &e = elements.back();
    for(int i = 0; i <= dim_of_elements; ++i) e.v[i] = v[i];
  }

  //determine_coordinate_extrema
  void determine_coordinate_extrema(Vertex &min_value, Vertex &max_value) const{
    for(int p = 0; p < dim_of_world; ++p){
      max_value[p] = -9e100;
      min_value[p] = 9e100;
    }
    for(int i = 0, i_end = vertices.size(); i < i_end; ++i){
      for(int p = 0; p < dim_of_world; ++p){
        if(vertices[i][p] > max_value[p]) max_value[p] = vertices[i][p];
        if(vertices[i][p] < min_value[p]) min_value[p] = vertices[i][p];
      }
    }
  }
};

}} //end namespace Dune::CurvedGeo

#endif //DUNE_CURVED_SURFACE_GRID_DATASTRUCTURES_HH
