#ifndef DUNE_CURVED_SURFACE_GRID_VERTEXMAP_HH
#define DUNE_CURVED_SURFACE_GRID_VERTEXMAP_HH

#include <dune/curvedsurfacegrid/surfacedistance/datastructures.hh>

#define VERTEXMAP_DEBUG_MODE 0

namespace Dune{ namespace CGeo{

//--------------------< VertexMap >
template<typename T> struct VertexMap{
  static int partition_threshold;
  static int depth_threshold_1d;
  static int depth_threshold_2d;
  static int depth_threshold_3d;
  static double config_epsilon;
  int dim_of_world;
  int depth;
  Vertex c_begin, c_mid, c_end; //coordinates
  VertexMap<T> *p0, *p1, *p2, *p3, *p4, *p5, *p6, *p7; //partitions
  std::vector<std::pair<Vertex, T> > entries; //VertexMap behaves like a map<Vertex, T>

  //methods
  VertexMap(const Vertex &c_begin, const Vertex &c_end, int dim_of_world, int depth=0):
                                dim_of_world(dim_of_world), depth(depth), c_begin(c_begin),
                          c_mid((c_begin + c_end) / 2.0), c_end(c_end), p0(NULL), p1(NULL),
                    p2(NULL), p3(NULL), p4(NULL), p5(NULL), p6(NULL), p7(NULL){/*$ACL_exclude*/}
  ~VertexMap();
  void clear();
  void split();
  void insert(const Vertex &v, const T &data);
  bool find(const Vertex &v, T &result){
    switch(dim_of_world){
      case 1: return find_1d(v, result);
      case 2: return find_2d(v, result);
      default: return find_3d(v, result);
    }
  }
  bool find_1d(const Vertex &v, T &result);
  bool find_2d(const Vertex &v, T &result);
  bool find_3d(const Vertex &v, T &result);
  #if VERTEXMAP_DEBUG_MODE != 0
    void debug_output_statistics();
    void debug_statistics_first_run(size_t &node_count, size_t &leaf_count,
                    size_t &crowded_leaf_count, size_t &entry_count, size_t &max_leaf_entries);
    void debug_statistics_second_run(double avg, size_t entry_count, double &std_dev_sum);
    void debug_output(int indentation=0);
  #endif
};

//                    ################################
//                    ##  VertexMap Implementation  ##
//                    ################################

template<typename T> int VertexMap<T>::partition_threshold = 64;
template<typename T> int VertexMap<T>::depth_threshold_1d = 19;
template<typename T> int VertexMap<T>::depth_threshold_2d = 10;
template<typename T> int VertexMap<T>::depth_threshold_3d = 7;
template<typename T> double VertexMap<T>::config_epsilon = epsilon;

//--------------------< destructor >
template<typename T> VertexMap<T>::~VertexMap(){ //$ACL_exclude
  if(p0 != NULL){
    delete p0;
    delete p1;
    if(dim_of_world >= 2){
      delete p2;
      delete p3;
      if(dim_of_world == 3){
        delete p4;
        delete p5;
        delete p6;
        delete p7;
      }
    }
  }
}

//--------------------< clear >
template<typename T> void VertexMap<T>::clear(){
  entries.clear();
  if(p0 != NULL){
    delete p0; p0 = NULL;
    delete p1; p1 = NULL;
    if(dim_of_world >= 2){
      delete p2; p2 = NULL;
      delete p3; p3 = NULL;
      if(dim_of_world == 3){
        delete p4; p4 = NULL;
        delete p5; p5 = NULL;
        delete p6; p6 = NULL;
        delete p7; p7 = NULL;
      }
    }
  }
}

//--------------------< split >
template<typename T> void VertexMap<T>::split(){
  Vertex c0 = c_begin, c1 = c_mid;
  p0 = new VertexMap(c0, c1, dim_of_world, depth + 1);
  c0.c[0] = c_mid.c[0];
  c1.c[0] = c_end.c[0];
  p1 = new VertexMap(c0, c1, dim_of_world, depth + 1);
  if(dim_of_world >= 2){
    c0.c[1] = c_mid.c[1];
    c1.c[1] = c_end.c[1];
    p3 = new VertexMap(c0, c1, dim_of_world, depth + 1);
    c0.c[0] = c_begin.c[0];
    c1.c[0] = c_mid.c[0];
    p2 = new VertexMap(c0, c1, dim_of_world, depth + 1);
    if(dim_of_world == 3){
      c0.c[2] = c_mid.c[2];
      c1.c[2] = c_end.c[2];
      p6 = new VertexMap(c0, c1, dim_of_world, depth + 1);
      c0.c[1] = c_begin.c[1];
      c1.c[1] = c_mid.c[1];
      p4 = new VertexMap(c0, c1, dim_of_world, depth + 1);
      c0.c[0] = c_mid.c[0];
      c1.c[0] = c_end.c[0];
      p5 = new VertexMap(c0, c1, dim_of_world, depth + 1);
      p7 = new VertexMap(c_mid, c_end, dim_of_world, depth + 1);
    }
  }
  for(typename std::vector<std::pair<Vertex, T> >::iterator it = entries.begin(),
                                             end = entries.end(); it != end; ++it){
    insert(it->first, it->second);
  }
  entries.clear();
}

//--------------------< insert >
template<typename T> void VertexMap<T>::insert(const Vertex &v, const T &data){
  if(dim_of_world == 3){ //3D-world
    if(p0 == NULL){
      entries.push_back(std::pair<Vertex, T>(v, data));
      if((int)entries.size() > partition_threshold && depth < depth_threshold_3d) split();
    }else{
      if(v.c[2] <= c_mid.c[2]){ //p0, p1, p2, p3
        if(v.c[1] <= c_mid.c[1]){ //p0, p1
          if(v.c[0] <= c_mid.c[0]) p0->insert(v, data);
          else p1->insert(v, data);
        }else{ //p2, p3
          if(v.c[0] <= c_mid.c[0]) p2->insert(v, data);
          else p3->insert(v, data);
        }
      }else{ //p4, p5, p6, p7
        if(v.c[1] <= c_mid.c[1]){ //p4, p5
          if(v.c[0] <= c_mid.c[0]) p4->insert(v, data);
          else p5->insert(v, data);
        }else{ //p6, p7
          if(v.c[0] <= c_mid.c[0]) p6->insert(v, data);
          else p7->insert(v, data);
        }
      }
    }
  }else if(dim_of_world == 2){ //2D-world
    if(p0 == NULL){
      entries.push_back(std::pair<Vertex, T>(v, data));
      if((int)entries.size() > partition_threshold && depth < depth_threshold_2d) split();
    }else{
      if(v.c[1] <= c_mid.c[1]){ //p0, p1
        if(v.c[0] <= c_mid.c[0]) p0->insert(v, data);
        else p1->insert(v, data);
      }else{ //p2, p3
        if(v.c[0] <= c_mid.c[0]) p2->insert(v, data);
        else p3->insert(v, data);
      }
    }
  }else{ //1D-world
    if(p0 == NULL){
      entries.push_back(std::pair<Vertex, T>(v, data));
      if((int)entries.size() > partition_threshold && depth < depth_threshold_1d) split();
    }else{
      if(v.c[0] <= c_mid.c[0]) p0->insert(v, data);
      else p1->insert(v, data);
    }
  }
}

//--------------------< find_1d >
template<typename T> bool VertexMap<T>::find_1d(const Vertex &v, T &result){
  if(p0 == NULL){
    for(typename std::vector<std::pair<Vertex, T> >::iterator it = entries.begin(),
                                               end = entries.end(); it != end; ++it){
      Vertex &t = it->first;
      if(abs(t.c[0] - v.c[0]) <= config_epsilon){
        result = it->second;
        return true;
      }
    }
  }else{
    if(v.c[0] - config_epsilon <= c_mid.c[0]) if(p0->find_1d(v, result)) return true;
    if(v.c[0] + config_epsilon >= c_mid.c[0]) if(p1->find_1d(v, result)) return true;
  }
  return false;
}

//--------------------< find_2d >
template<typename T> bool VertexMap<T>::find_2d(const Vertex &v, T &result){
  if(p0 == NULL){
    for(typename std::vector<std::pair<Vertex, T> >::iterator it = entries.begin(),
                                               end = entries.end(); it != end; ++it){
      Vertex &t = it->first;
      if(abs(t.c[0] - v.c[0]) <= config_epsilon && abs(t.c[1] - v.c[1]) <= config_epsilon){
        result = it->second;
        return true;
      }
    }
  }else{
    if(v.c[1] - config_epsilon <= c_mid.c[1]){ //p0, p1
      if(v.c[0] - config_epsilon <= c_mid.c[0]) if(p0->find_2d(v, result)) return true;
      if(v.c[0] + config_epsilon >= c_mid.c[0]) if(p1->find_2d(v, result)) return true;
    }
    if(v.c[1] + config_epsilon >= c_mid.c[1]){ //p2, p3
      if(v.c[0] - config_epsilon <= c_mid.c[0]) if(p2->find_2d(v, result)) return true;
      if(v.c[0] + config_epsilon >= c_mid.c[0]) if(p3->find_2d(v, result)) return true;
    }
  }
  return false;
}

//--------------------< find_3d >
template<typename T> bool VertexMap<T>::find_3d(const Vertex &v, T &result){
  if(p0 == NULL){
    for(typename std::vector<std::pair<Vertex, T> >::iterator it = entries.begin(),
                                               end = entries.end(); it != end; ++it){
      Vertex &t = it->first;
      if(abs(t.c[0] - v.c[0]) <= config_epsilon && abs(t.c[1] - v.c[1]) <= config_epsilon
                                                 && abs(t.c[2] - v.c[2]) <= config_epsilon){
        result = it->second;
        return true;
      }
    }
  }else{
    if(v.c[2] - config_epsilon <= c_mid.c[2]){ //p0, p1, p2, p3
      if(v.c[1] - config_epsilon <= c_mid.c[1]){ //p0, p1
        if(v.c[0] - config_epsilon <= c_mid.c[0]) if(p0->find_3d(v, result)) return true;
        if(v.c[0] + config_epsilon >= c_mid.c[0]) if(p1->find_3d(v, result)) return true;
      }
      if(v.c[1] + config_epsilon >= c_mid.c[1]){ //p2, p3
        if(v.c[0] - config_epsilon <= c_mid.c[0]) if(p2->find_3d(v, result)) return true;
        if(v.c[0] + config_epsilon >= c_mid.c[0]) if(p3->find_3d(v, result)) return true;
      }
    }
    if(v.c[2] + config_epsilon >= c_mid.c[2]){ //p4, p5, p6, p7
      if(v.c[1] - config_epsilon <= c_mid.c[1]){ //p4, p5
        if(v.c[0] - config_epsilon <= c_mid.c[0]) if(p4->find_3d(v, result)) return true;
        if(v.c[0] + config_epsilon >= c_mid.c[0]) if(p5->find_3d(v, result)) return true;
      }
      if(v.c[1] + config_epsilon >= c_mid.c[1]){ //p6, p7
        if(v.c[0] - config_epsilon <= c_mid.c[0]) if(p6->find_3d(v, result)) return true;
        if(v.c[0] + config_epsilon >= c_mid.c[0]) if(p7->find_3d(v, result)) return true;
      }
    }
  }
  return false;
}

//--------------------< debug_output_statistics >
#if VERTEXMAP_DEBUG_MODE != 0
template<typename T> void VertexMap<T>::debug_output_statistics(){
  std::cout << "\nVertexMap statistics:" << std::endl;
  size_t node_count = 0, leaf_count = 0, crowded_leaf_count = 0, entry_count = 0,
                                                              max_leaf_entries = 0;
  debug_statistics_first_run(node_count, leaf_count, crowded_leaf_count,
                                           entry_count, max_leaf_entries);
  double avg = (double)entry_count / leaf_count;
  double std_dev = 0.0;
  if(entry_count > 0){
    debug_statistics_second_run(avg, entry_count, std_dev);
    std_dev = sqrt(std_dev);
  }
  std::cout << "  node-count:         " << node_count << std::endl;
  std::cout << "  leaf-count:         " << leaf_count
                            << " (" << (double)leaf_count / node_count * 100.0 << "%)" << std::endl;
  std::cout << "  crowded leaf-count: " << crowded_leaf_count
                    << " (" << (double)crowded_leaf_count / leaf_count * 100.0 << "%)" << std::endl;
  std::cout << "  entry-count:        " << entry_count << std::endl;
  std::cout << "  max. leaf-entries:  " << max_leaf_entries << std::endl;
  std::cout << "  avg. leaf-entries:  " << avg << std::endl;
  std::cout << "  std. deviation:     " << std_dev << std::endl;
}
#endif

//--------------------< debug_statistics_first_run >
#if VERTEXMAP_DEBUG_MODE != 0
template<typename T> void VertexMap<T>::debug_statistics_first_run(size_t &node_count,
     size_t &leaf_count, size_t &crowded_leaf_count, size_t &entry_count, size_t &max_leaf_entries){
  ++node_count;
  if(p0 == NULL){
    ++leaf_count;
    entry_count += entries.size();
    max_leaf_entries = std::max(max_leaf_entries, entries.size());
    if(entries.size() > partition_threshold) ++crowded_leaf_count;
  }else{
    p0->debug_statistics_first_run(node_count, leaf_count,
                crowded_leaf_count, entry_count, max_leaf_entries);
    p1->debug_statistics_first_run(node_count, leaf_count,
                crowded_leaf_count, entry_count, max_leaf_entries);
    if(dim_of_world >= 2){
      p2->debug_statistics_first_run(node_count, leaf_count,
                crowded_leaf_count, entry_count, max_leaf_entries);
      p3->debug_statistics_first_run(node_count, leaf_count,
                crowded_leaf_count, entry_count, max_leaf_entries);
      if(dim_of_world == 3){
        p4->debug_statistics_first_run(node_count, leaf_count,
                crowded_leaf_count, entry_count, max_leaf_entries);
        p5->debug_statistics_first_run(node_count, leaf_count,
                crowded_leaf_count, entry_count, max_leaf_entries);
        p6->debug_statistics_first_run(node_count, leaf_count,
                crowded_leaf_count, entry_count, max_leaf_entries);
        p7->debug_statistics_first_run(node_count, leaf_count,
                crowded_leaf_count, entry_count, max_leaf_entries);
      }
    }
  }
}
#endif

//--------------------< debug_statistics_second_run >
#if VERTEXMAP_DEBUG_MODE != 0
template<typename T> void VertexMap<T>::debug_statistics_second_run(double avg,
                                        size_t entry_count, double &std_dev_sum){
  if(p0 == NULL){
    double deviation = (double)entries.size() - avg;
    std_dev_sum += (deviation * deviation) / entry_count;
  }else{
    p0->debug_statistics_second_run(avg, entry_count, std_dev_sum);
    p1->debug_statistics_second_run(avg, entry_count, std_dev_sum);
    if(dim_of_world >= 2){
      p2->debug_statistics_second_run(avg, entry_count, std_dev_sum);
      p3->debug_statistics_second_run(avg, entry_count, std_dev_sum);
      if(dim_of_world == 3){
        p4->debug_statistics_second_run(avg, entry_count, std_dev_sum);
        p5->debug_statistics_second_run(avg, entry_count, std_dev_sum);
        p6->debug_statistics_second_run(avg, entry_count, std_dev_sum);
        p7->debug_statistics_second_run(avg, entry_count, std_dev_sum);
      }
    }
  }
}
#endif

//--------------------< debug_output >
#if VERTEXMAP_DEBUG_MODE != 0
template<typename T> void VertexMap<T>::debug_output(int indentation/*=0*/){
  std::string id_str = std::string(indentation, ' ');
  if(indentation == 0) std::cout << "\nVertexMap:" << std::endl;
  if(p0 == NULL){
    for(typename std::vector<std::pair<Vertex, T> >::iterator it = entries.begin(),
                                               end = entries.end(); it != end; ++it){
      std::cout << id_str << it->first << " -> " << it->second << std::endl;
    }
  }else{
    std::cout << id_str << "p0{" << std::endl;
    p0->debug_output(indentation + 2);
    std::cout << id_str << "}" << std::endl;
    std::cout << id_str << "p1{" << std::endl;
    p1->debug_output(indentation + 2);
    std::cout << id_str << "}" << std::endl;
    if(dim_of_world >= 2){
      std::cout << id_str << "p2{" << std::endl;
      p2->debug_output(indentation + 2);
      std::cout << id_str << "}" << std::endl;
      std::cout << id_str << "p3{" << std::endl;
      p3->debug_output(indentation + 2);
      std::cout << id_str << "}" << std::endl;
      if(dim_of_world == 3){
        std::cout << id_str << "p4{" << std::endl;
        p4->debug_output(indentation + 2);
        std::cout << id_str << "}" << std::endl;
        std::cout << id_str << "p5{" << std::endl;
        p5->debug_output(indentation + 2);
        std::cout << id_str << "}" << std::endl;
        std::cout << id_str << "p6{" << std::endl;
        p6->debug_output(indentation + 2);
        std::cout << id_str << "}" << std::endl;
        std::cout << id_str << "p7{" << std::endl;
        p7->debug_output(indentation + 2);
        std::cout << id_str << "}" << std::endl;
      }
    }
  }
}
#endif

}} //end namespace Dune::CurvedGeo

#endif //DUNE_CURVED_SURFACE_GRID_VERTEXMAP_HH
